#!/bin/csh -f
#set echo verbose

# /usr/local/bin/fpromisc
#
# Copyright 2000, 2003 Bob Toxen.  All Rights Reserved.
#
# Purchasers of my book "Real World Linux Security, 2nd ed" may copy this
# script as needed to install and use on any system that they
# administer.  Others may not copy or use it without obtaining
# specific written permission by contacting the author at
# book@cavu.com.
#
# Find processes that have sockets open for other than TCP/IP,
# UDP/IP, or unix protocols and thus are candidates for being
# rogues having an interface open in promiscuous mode.
#
# Exit code 0 if no candidates or -1/255 on error or #candidates
#
# Unfortunately the rogues cannot be separated from legitimate
# processes such as those using pipes on pre-2.2 kernels.  The
# latter can be guessed by their having sequential PIDs and a
# common PPID but these still could be rogues.
#
# Recommended procedure:
#   1. Run fpromisc, storing output as a "snapshot"
#   2. Run promiscalert (if it detects problem, it will
#      invoke promisc and email the results)

umask 077
set noclobber
set files=""
set tmpdir="/tmp/fprom$$"
set fpid="$tmpdir/P"
set fpid2="$tmpdir/Pb"
set fpidbad="$tmpdir/Pc"
set ftcp="$tmpdir/T"
set fudp="$tmpdir/U"
set funix="$tmpdir/X"
set finum="$tmpdir/I"

# Make sure we are root
if ( ! -o /bin/sh ) then
	echo "${0}: Must be root"
	exit 255
endif

# Create our own dir for safety
mkdir $tmpdir
if ( $status != 0 ) then
	echo "${0}: Cannot mkdir $tmpdir"
	exit 255
endif

# Gather all info now to reduce mismatches from out-of-date
if ( -e $fpid ) goto exists
set files="$files $fpid"
file /proc/[0-9]*/fd/* > $fpid
if ( $status != 0 ) then
	echo "${0}: Cannot use file to find processes"
	goto cleanup
endif

if ( -e $ftcp ) goto exists
set files="$files $ftcp"
cat /proc/net/tcp > $ftcp
if ( $status != 0 ) then
	echo "${0}: Cannot cp tcp"
	goto cleanup
endif

if ( -e $fudp ) goto exists
set files="$files $fudp"
cat /proc/net/udp > $fudp
if ( $status != 0 ) then
	echo "${0}: Cannot cp udp"
	goto cleanup
endif

if ( -e $funix ) goto exists
set files="$files $funix"
cat /proc/net/unix > $funix
if ( $status != 0 ) then
	echo "${0}: Cannot cp unix"
	goto cleanup
endif

# Chop PID lines to applicable PIDs and INUMs (kernel dep.)

fgrep -q 'to socket:' $fpid
if ( $status != 0 ) goto pre22kernel

if ( -e $fpid2 ) goto exists
set files="$files $fpid2"
fgrep 'to socket:' $fpid \
  | sed -e 's,/proc/,,' -e 's,/.*socket:\[,	,' -e 's,],,' \
  | sort +1 > $fpid2
if ( $status != 0 ) then
	echo "${0}: Cannot fgrep"
	goto cleanup
endif
goto gotpids

pre22kernel:
fgrep -q 'to [0000]:' $fpid
if ( $status != 0 ) then
	echo "${0}: No open sockets: something probably is broken"
	echo Did: "fgrep -q 'to [0000]:' $fpid"
	echo more $fpid :
	more $fpid
	goto cleanup
endif
if ( -e $fpid2 ) goto exists
set files="$files $fpid2"
fgrep 'to [0000]:' $fpid | sed -e 's,/proc/,,' -e 's,/.*:,	,' \
  | sort +1 > $fpid2
if ( $status != 0 ) then
	echo "${0}: Cannot fgrep"
	goto cleanup
endif

gotpids:

# Generate list of ok socket inums
if ( -e $finum ) goto exists
set files="$files $finum"

# Format of /proc/net/{tcp,udp,unix} changed in 2.4
uname -r | grep -q '^2\.2'
if ( $status != 0 ) then
		# 2.4
	((tail +2 $ftcp;tail +2 $fudp) \
	  | sed -e 's/.*:[^ ]* [^ ]*  *[^ ]*  *[^ ]*  *\([^ ][^ ]*\) .*/\1/'; \
	  tail +2 $funix | sed -e 's, /.*,,' -e 's,.* ,,') | sort > $finum
else
		# 2.2 (we even handle 2.0 elsewhere)
	(tail +2 $ftcp;tail +2 $fudp;tail +2 $funix | sed 's, /.*,,') \
	  | sed -e 's/  *$//' -e 's/.* //' | sort > $finum
endif

# Give stats (for sanity):
echo "normal     unique open network inums: `wc -l < $finum`"
echo "total  NON-unique open network inums: `wc -l < $fpid2`"

# Find PIDs whose inums are not tcp or udp or unix
# -v 1: 	only output lines in file 1 (inums) that cannot be joined
# -j1 2:	join file 1 along field 2 (default is 1)
# -t'<tab>'	Tab is field separator
# NOTE: -t'<tab>'
# NOTE: join cannot be in backprimes: tab would be trashed
if ( -e $fpidbad ) goto exists
set files="$files $fpidbad"
join -v 1 -j1 2 -j2 1 -t'	' $fpid2 $finum \
  | sed 's/.*	//' | sort -un > $fpidbad
set bad=(`cat $fpidbad`)

if ( "$files" != "" ) /bin/rm $files
# We do not want to do a /bin/rm -rf due to danger of fubar
rmdir $tmpdir

echo $#bad suspicious processes: $bad
if ( $#bad > 0 ) then
	ps lww $bad
# Return 1 on suspicious PIDs; 0 otherwise
	exit $#bad
endif

exit 0

exists:
	echo "${0}: File exists but should not"
	goto cleanup

cleanup:
	if ( "$files" != "" ) /bin/rm $files
		# We do not want to do a /bin/rm -rf due to danger of fubar
	rmdir $tmpdir
	exit 255
