/*
 * ports: indicate open ports on a Linux system; may work on some
 * Unix boxes.  Note that root access is not required.
 *
 * May need to be tweaked for different kernel versions that may use
 * slightly different syntax in their /proc/net/{tcp,upd} files.
 * Validated on Red Hat 7.3 and Mandrake 8.2.
 *
 * Copyright 2001, 2002 Bob Toxen.  All Rights Reserved.
 *
 * Purchasers of the book "Real World Linux Security --
 * Intrusion Prevention, Detection and Recovery" may copy this
 * script as needed to install and use on up to 10 systems that they
 * administer.  Others may not copy or use it without obtaining
 * specific written permission by contacting the author at
 * book@cavu.com.
 *
 * Offered as is with no warranty of any kind.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

		/* data[]: {lcl,remote}{addr,port} state inum. */
#define	D_LADDR		0
#define	D_LPORT		1
#define	D_RADDR		2
#define	D_RPORT		3
#define	D_STATE		4
#define	D_INUM		5

#define	TCPIDX		0
#define	UDPIDX		1

#define	LINEMAX		512	/* Plenty big. */

const
char	*file[] = {
	"/proc/net/tcp",
	"/proc/net/udp",
};
const
char	*proto[] = {
	"TCP",
	"UDP",
};
				/* Proto names in lower-case. */
const
char	*protol[] = {
	"tcp",
	"udp",
};
int	data[6];	/* {lcl,remote}{addr,port} state inum. */
int	alerts;
char	line[LINEMAX];
union	{
	struct	in_addr	str;
	int		i;
} aconv;
const
char	*look(int idx);
int	hiport(int p);
int	badport(int p);
void	putport(int port, int idx);

/* Adapted from linux/tcp.h due to include file conflicts in RH. */
enum {
	TCP_ESTABLISHED = 1,
	TCP_SYN_SENT,
	TCP_SYN_RECV,
	TCP_FIN_WAIT1,
	TCP_FIN_WAIT2,
	TCP_TIME_WAIT,
	TCP_CLOSE,
	TCP_CLOSE_WAIT,
	TCP_LAST_ACK,
	TCP_LISTEN,
	TCP_CLOSING		/* now a valid state */
};

/* Adapted from linux/net.h due to include file conflicts in RH. */
typedef enum {
	SS_FREE = 0,		/* not allocated		*/
	SS_UNCONNECTED,		/* unconnected to any socket	*/
	SS_CONNECTING,		/* in process of connecting	*/
	SS_CONNECTED,		/* connected to socket		*/
	SS_DISCONNECTING	/* in process of disconnecting	*/
} socket_state;

struct	statestr {
	int	num;
	char	*name;
} const state[] = {
	{TCP_ESTABLISHED,	"ESTABLISD"},
	{TCP_SYN_SENT,		"SYN_SENT"},
	{TCP_SYN_RECV,		"SYN_RECV"},
	{TCP_FIN_WAIT1,		"FIN_WAIT1"},
	{TCP_FIN_WAIT2,		"FIN_WAIT2"},
	{TCP_TIME_WAIT,		"TIME_WAIT"},
	{TCP_CLOSE,		"CLOSE"},
	{TCP_CLOSE_WAIT,	"CLOSE_WAIT"},
	{TCP_LAST_ACK,		"LAST_ACK"},
	{TCP_LISTEN,		"LISTEN"},
	{TCP_CLOSING,		"CLOSING"},
	{SS_FREE,		"SS_FREE"},
	{SS_UNCONNECTED,	"SS_UNCONNECTED"},
	{SS_CONNECTING,		"SS_CONNECTING"},
	{SS_CONNECTED,		"SS_CONNECTED"},
	{SS_DISCONNECTING,	"SS_DISCONNECTING"},
};

/* Known ports, some evil, not always in /etc/services. */
struct	portstr	{
	const
	char	*name;
	int	num;
	int	idx;
} ports[] = {
/*	name			port	protocol */
	{"zero",		0,	TCPIDX},
	{"zero",		0,	UDPIDX},

	{"x",			6000,	TCPIDX},
	{"x+1",			6001,	TCPIDX},
	{"x+2",			6002,	TCPIDX},
	{"x+3",			6003,	TCPIDX},
	{"x+4",			6004,	TCPIDX},
	{"x+5",			6005,	TCPIDX},
	{"x+6",			6006,	TCPIDX},
	{"x+7",			6007,	TCPIDX},
	{"x+8",			6008,	TCPIDX},
	{"x+9",			6009,	TCPIDX},
	{"x+ssh",		6010,	TCPIDX},
	{"x+ssh+1",		6011,	TCPIDX},
	{"x+ssh+2",		6012,	TCPIDX},
	{"x+ssh+3",		6013,	TCPIDX},
	{"x+ssh+4",		6014,	TCPIDX},
	{"xfont",		7100,	TCPIDX},

	{"ftpdata",		20,	TCPIDX},
	{"ssh",			22,	TCPIDX},
	{"pop3",		110,	TCPIDX},
	{"sunrpc",		111,	UDPIDX},
	{"imap2",		143,	TCPIDX},
	{"snmp",		161,	UDPIDX},
	{"route",		520,	UDPIDX},
	{"socks",		1080,	TCPIDX},

	{"mount",		635,	UDPIDX},
	{"nfs",			2049,	TCPIDX},
	{"nfs",			2049,	UDPIDX},

	{"http",		80,	TCPIDX},
	{"https",		443,	TCPIDX},

	{"dtk",			365,	TCPIDX},
	{"dtku",		365,	UDPIDX},

	{"snews",		563,	TCPIDX},

	{"ingresslock",		1524,	TCPIDX},

	{"squid",		3128,	TCPIDX},

	{"smbd",		1048,	UDPIDX},
	{"netbios-ns",		137,	TCPIDX},
	{"netbios-dgm",		138,	TCPIDX},
	{"netbios-ssn",		139,	TCPIDX},
	{"netbios-ns",		137,	UDPIDX},
	{"netbios-dgm",		138,	UDPIDX},
	{"netbios-ssn",		139,	UDPIDX},

	{"WRAPPERTROJ",		421,	TCPIDX}, /* TCP Wrappers Trojan. */
	{"WRAPPERTROJU",	421,	UDPIDX}, /* UDP Wrappers Trojan. */

	{"KAZAA",		1214,	TCPIDX},
	{"SUBSEVEN",		1243,	TCPIDX},
	{"SUBSEVEN2",		1234,	TCPIDX},
	{"SUBSEVEN3",		27374,	TCPIDX},

	{"DEEPTHROAT",		2140,	UDPIDX},

	{"SOCDIMI",		31337,	TCPIDX},
	{"BACKORIFICE",		31337,	UDPIDX},

	{"MSTREAM_M1",		6838,	UDPIDX},
	{"MSTREAM_M2",		9325,	UDPIDX},
	{"MSTREAM_Z1",		10498,	UDPIDX},
	{"MSTREAM_Z2",		7983,	UDPIDX},

	{"PORTD",		2400,	TCPIDX},

	{"HSH002",		7890,	TCPIDX},

	{"TRINOO",		27444,	UDPIDX},
	{"WINDOWSTRINOO",	34555,	UDPIDX},
	{"WINDOWSTRINOOCLIENT",	35555,	UDPIDX},

	{"BARBEDWIRE1",		16660,	TCPIDX},
	{"BARBEDWIRE1",		16660,	UDPIDX},
	{"BARBEDWIRE2",		60001,	TCPIDX},
	{"BARBEDWIRE2",		60001,	UDPIDX},
	{"BARBEDWIRE_AGENT",	65000,	TCPIDX},
	{"BARBEDWIRE_AGENT",	65000,	UDPIDX},
};

/* Some known cracker ports: add your own. */
int	badports[] = {
	421,
	1214,
	1234,
	1243,
	1524,
	2140,
	2400,
	6838,
	7890,
	7983,
	9325,
	10498,
	16660,
	27374,
	27444,
	31337,
	34555,
	35555,
	60001,
	65000,
};

main(int argc, char **argv)
{
	char	*s;
	char	*t;
	int	i;
	int	j;
	FILE	*fp;
	static
	int	ret;
	struct	hostent	*hostptr;
	const
	char	*alert;

	argc--;
	argv++;

			/* Ensure same #ele in file & proto. */
	if (sizeof file/sizeof *file
	  != sizeof proto/sizeof *proto
	  || sizeof proto/sizeof *proto
	  != sizeof protol/sizeof *protol)
		fprintf(stderr, "bug\n"), exit(1);

			/* Process each pseudo file. */
	for (i=0; i<sizeof file/sizeof *file; i++) {
		fp = fopen(file[i], "r");
		if (!fp)
			perror(file[i]), exit(2);

			/* Skip first line (header). */
		if (!fgets(line, LINEMAX, fp) || !strchr(line, '\n'))
			fprintf(stderr, "syntax error\n"), exit(3);

		if (i)
			putchar('\n');
		printf("%s\n", proto[i]);
		printf(" Lcl port       Rmt port     Status"
		  "       Rmt IP          Rmt host\n");

			/* Process each line of data. */
		for (;;) {
			if (!fgets(line, LINEMAX, fp))
				break;
			s = strchr(line, '\n');
			if (!s) {
				fprintf(stderr,
				  "%s: line too long\n", file[i]);
				if (!ret)
					ret = 3 + i;
				break;
			}

					/* Strip trailing spaces. */
			while (s > line && s[-1] == ' ')
				s--;

			*s = 0;		/* Truncate string. */
			alert = 0;	/* Init. alert flag.*/

			s = strchr(line, ':');
			for (j=0; j<5; j++) {
				if (!s) {
synerr:;
					fprintf(stderr,
					  "%s: Syntax error\n",
					  proto[i]);
					exit(5 + i);
				}
					/* Skip colon and spaces. */
				while (*s == ' ' || *s == ':')
					s++;
				data[j] = strtoul(s, &t, 16);
				if (!t || !*t)
					goto synerr;
				s = t;
			}
					/* Find inum (last field). */
			s = strrchr(s, ' ');
			if (!s)
				goto synerr;
			while (*s == ' ')
				s++;
			data[D_INUM] = atoi(s);
					/* Flag suspicious ports. */
			alert = look(i);
			if (alert)
				putchar('*');
			else
				putchar(' ');

					/* Local & rmt ports, state. */
			putport(data[D_LPORT], i);
			putport(data[D_RPORT], i);
			printf("%02X", data[4]);
			for (j=0; j<sizeof state/sizeof *state; j++)
				if (state[j].num == data[D_STATE]) {
					printf("=%-10s",
					  state[j].name);
					break;
				}
			if (j >= sizeof state/sizeof *state)
				printf("          ");

					/* Remote host & IP. */
			aconv.i = data[D_RADDR];
			s = inet_ntoa(aconv.str);
			if (!s)
				s = "error";
			if (data[D_RADDR]) {
				sethostent(1);
				hostptr = gethostbyaddr(
				  (char *)&data[D_RADDR],
				  4, AF_INET);
				if (hostptr)
					t = hostptr->h_name;
				else {
					t = "IP?";
					herror(s);
				}
			} else
				t = "local";
			printf("%-15s %s\n", s, t);
			if (alert) {
				printf("%s\n\n", alert);
				alerts++;
			}
		}
		fclose(fp);
	}

	if (alerts) {
		printf("\n%d alerts: possible security breach\n",
		  alerts);
		exit(-1);
	}
	exit(0);
}

/* Print port number and service name. */
void
putport(int port, int idx)
{
	struct	servent	*srv;
	const
	char	*s	= 0;
	int	i;

		/*
		 * Priority to internal names
		 * in case /etc/services altered.
		 */
	for (i=0; i<sizeof ports/sizeof *ports; i++)
		if (ports[i].num == port
		  && ports[i].idx == idx) {
			s = ports[i].name;
			break;
		}
	if (!s) {
		setservent(1);
		srv = getservbyport(htons(port), protol[idx]);
		if (srv)
			s = srv->s_name;
	}
	if (!s)
		s = "?";
	printf("%5d=%-7s ", port, s);
}

/* Flag suspicious ports: add your own tests. */
const
char *
look(int idx)
{
	const
	char	*alert	= 0;

	if (badport(data[D_LPORT]))
		alert = "*** cracker server";
	else if (badport(data[D_RPORT]))
		alert = "*** cracker client";
	else if (hiport(data[D_LPORT])) {
		if ((idx == 0 && data[D_STATE] == TCP_LISTEN
		  || idx == 1 && data[D_STATE] == TCP_CLOSE))
			alert = "*** High numbered port listening";
		else if (idx == 0 && hiport(data[D_RPORT]))
			alert = "*** High numbered ports communing";

	}
	return alert;
}

/* True if high port number without standard service. */
/* X+SSH normally uses port 6010 for the first connection. */
hiport(int p)
{
	return p >= 1024 && p != 1080 && p != 2049
	  && p != 6000 && p != 6010
	  && p != 3128;
}

/* True if known cracker server port. */
badport(int p)
{
	int	i;

	for (i=0; i<sizeof badports/sizeof *badports; i++)
		if (badports[i] == p)
			return 1;
	return 0;
}
